from binance.lib.utils import get_uuid


def user_data_start(self, id=None):
    """Start user data stream (USER_STREAM)

    Args:
        id (str): A unique id for the request

    Message sent:

    .. code-block:: json

        {
            "id": "d3df8a61-98ea-4fe0-8f4e-0fcea5d418b0",
            "method": "userDataStream.start",
            "params": {
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A"
            }
        }

    Response:

    .. code-block:: json

        {
            "id": "d3df8a61-98ea-4fe0-8f4e-0fcea5d418b0",
            "status": 200,
            "result": {
                "listenKey": "xs0mRXdAKlIPDRFrlPcw0qI41Eh3ixNntmymGyhrhgqo7L6FuLaWArTD7RLP"
            },
            "rateLimits": [
                {
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 1
                }
            ]
        }


    """

    if not id:
        id = get_uuid()

    payload = {
        "id": id,
        "method": "userDataStream.start",
        "params": {"apiKey": self.api_key},
    }

    self.send(payload)


def user_data_ping(self, listenKey: str, id=None):
    """Keepalive a user data stream to prevent a time out.

    Args:
        listenKey (str): The listen key from the user data stream
        id (str): A unique id for the request

    Message sent:

    .. code-block:: json

        {
            "id": "815d5fce-0880-4287-a567-80badf004c74",
            "method": "userDataStream.ping",
            "params": {
                "listenKey": "xs0mRXdAKlIPDRFrlPcw0qI41Eh3ixNntmymGyhrhgqo7L6FuLaWArTD7RLP",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A"
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "815d5fce-0880-4287-a567-80badf004c74",
            "status": 200,
            "response": {},
            "rateLimits": [
                {
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 1
                }
            ]
        }


    """

    if not id:
        id = get_uuid()

    payload = {
        "id": id,
        "method": "userDataStream.ping",
        "params": {"apiKey": self.api_key, "listenKey": listenKey},
    }

    self.send(payload)


def user_data_stop(self, listenKey: str, id=None):
    """Stop user data stream

    Args:
        listenKey (str): The listen key from the user data stream
        id (str): A unique id for the request

    Message sent:

    .. code-block:: json

        {
            "id": "819e1b1b-8c06-485b-a13e-131326c69599",
            "method": "userDataStream.stop",
            "params": {
                "listenKey": "xs0mRXdAKlIPDRFrlPcw0qI41Eh3ixNntmymGyhrhgqo7L6FuLaWArTD7RLP",
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A"
            }
        }

    Response:

    .. code-block:: json

        {
            "id": "819e1b1b-8c06-485b-a13e-131326c69599",
            "status": 200,
            "response": {},
            "rateLimits": [
                {
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 1
                }
            ]
        }

    """

    if not id:
        id = get_uuid()

    payload = {
        "id": id,
        "method": "userDataStream.stop",
        "params": {"apiKey": self.api_key, "listenKey": listenKey},
    }

    self.send(payload)



import json
import time
import uuid
from collections import OrderedDict
from urllib.parse import urlencode
from binance.lib.authentication import hmac_hashing
from binance.error import (
    ParameterRequiredError,
    ParameterValueError,
    ParameterTypeError,
    WebsocketClientError,
)


def cleanNoneValue(d) -> dict:
    out = {}
    for k in d.keys():
        if d[k] is not None:
            out[k] = d[k]
    return out


def check_required_parameter(value, name):
    if not value and value != 0:
        raise ParameterRequiredError([name])


def check_required_parameters(params):
    """Validate multiple parameters
    params = [
        ['btcusdt', 'symbol'],
        [10, 'price']
    ]

    """
    for p in params:
        check_required_parameter(p[0], p[1])


def check_enum_parameter(value, enum_class):
    if value not in set(item.value for item in enum_class):
        raise ParameterValueError([value])


def check_type_parameter(value, name, data_type):
    if value is not None and type(value) != data_type:
        raise ParameterTypeError([name, data_type])


def get_timestamp():
    return int(time.time() * 1000)


def encoded_string(query):
    return urlencode(query, True).replace("%40", "@")


def convert_list_to_json_array(symbols):
    if symbols is None:
        return symbols
    res = json.dumps(symbols)
    return res.replace(" ", "")


def config_logging(logging, logging_level, log_file: str = None):
    """Configures logging to provide a more detailed log format, which includes date time in UTC
    Example: 2021-11-02 19:42:04.849 UTC <logging_level> <log_name>: <log_message>

    Args:
        logging: python logging
        logging_level (int/str): For logging to include all messages with log levels >= logging_level. Ex: 10 or "DEBUG"
                                 logging level should be based on https://docs.python.org/3/library/logging.html#logging-levels
    Keyword Args:
        log_file (str, optional): The filename to pass the logging to a file, instead of using console. Default filemode: "a"
    """

    logging.Formatter.converter = time.gmtime  # date time in GMT/UTC
    logging.basicConfig(
        level=logging_level,
        filename=log_file,
        format="%(asctime)s.%(msecs)03d UTC %(levelname)s %(name)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def get_uuid():
    return str(uuid.uuid4())


def purge_map(map: map):
    """Remove None values from map"""
    return {k: v for k, v in map.items() if v is not None and v != "" and v != 0}


def websocket_api_signature(api_key: str, api_secret: str, parameters: dict):
    """Generate signature for websocket API
    Args:
        api_key (str): API key.
        api_secret (str): API secret.
        params (dict): Parameters.
    """

    if not api_key or not api_secret:
        raise WebsocketClientError(
            "api_key and api_secret are required for websocket API signature"
        )

    parameters["timestamp"] = get_timestamp()
    parameters["apiKey"] = api_key

    parameters = OrderedDict(sorted(parameters.items()))
    parameters["signature"] = hmac_hashing(api_secret, urlencode(parameters))

    return parameters



from binance.lib.utils import get_uuid, purge_map, websocket_api_signature


def account(self, **kwargs):
    """Account information (USER_DATA)

    Keyword Args:
        id (str, optional): Client generated ID.
        recvWindow (int, optional): The value cannot be greater than 60000

    Message sent:

    .. code-block:: json

        {
            "id": "605a6d20-6588-4cb9-afa0-b0ab087507ba",
            "method": "account.status",
            "params": {
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "83303b4a136ac1371795f465808367242685a9e3a42b22edb4d977d0696eb45c",
                "timestamp": 1660801839480
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "605a6d20-6588-4cb9-afa0-b0ab087507ba",
            "status": 200,
            "result": {
                "makerCommission": 15,
                "takerCommission": 15,
                "buyerCommission": 0,
                "sellerCommission": 0,
                "canTrade": true,
                "canWithdraw": true,
                "canDeposit": true,
                "commissionRates": {
                    "maker": "0.00150000",
                    "taker": "0.00150000",
                    "buyer": "0.00000000",
                    "seller":"0.00000000"
                },
                "brokered": false,
                "requireSelfTradePrevention": false,
                "updateTime": 1660801833000,
                "accountType": "SPOT",
                "balances": [
                    {
                        "asset": "BNB",
                        "free": "0.00000000",
                        "locked": "0.00000000"
                    },
                    {
                        "asset": "BTC",
                        "free": "1.3447112",
                        "locked": "0.08600000"
                    },
                    {
                        "asset": "USDT",
                        "free": "1021.21000000",
                        "locked": "0.00000000"
                    }
                ],
                "permissions": [
                    "SPOT"
                ]
            },
            "rateLimits": [
                {
                    "rateLimitType": "REQUEST_WEIGHT",
                    "interval": "MINUTE",
                    "intervalNum": 1,
                    "limit": 1200,
                    "count": 10
                }
            ]
        }

    """

    parameters = purge_map(kwargs)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "account.status",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)


def order_rate_limit(self, **kwargs):
    """Account order rate limits (USER_DATA)

    Keyword Args:
        id (str, optional): Client generated ID.
        recvWindow (int, optional): The value cannot be greater than 60000

    Message sent:

    .. code-block:: json

        {
            "id": "d3783d8d-f8d1-4d2c-b8a0-b7596af5a664",
            "method": "account.rateLimits.orders",
            "params": {
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "76289424d6e288f4dc47d167ac824e859dabf78736f4348abbbac848d719eb94",
                "timestamp": 1660801839500
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "d3783d8d-f8d1-4d2c-b8a0-b7596af5a664",
            "status": 200,
            "result": [{
                "rateLimitType": "ORDERS",
                "interval": "SECOND",
                "intervalNum": 1,
                "limit": 200
            },
            {
                "rateLimitType": "ORDERS",
                "interval": "DAY",
                "intervalNum": 1,
                "limit": 100000
            }],
            "rateLimits": [{
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 10
            }]
        }

    """

    parameters = purge_map(kwargs)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "account.rateLimits.orders",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)


def order_history(self, **kwargs):
    """Account order history (USER_DATA)

    Keyword Args:
        id (str, optional): Client generated ID.
        symbol (str, optional): Symbol
        orderId (int, optional): Order ID
        startTime (int, optional): Timestamp in ms
        endTime (int, optional): Timestamp in ms
        limit (int, optional): Default 500; max 1000.
        recvWindow (int, optional): The value cannot be greater than 60000

    Message sent:

    .. code-block:: json

        {
            "id": "734235c2-13d2-4574-be68-723e818c08f3",
            "method": "allOrders",
            "params": {
                "symbol": "BTCUSDT",
                "startTime": 1660780800000,
                "endTime": 1660867200000,
                "limit": 5,
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "f50a972ba7fad92842187643f6b930802d4e20bce1ba1e788e856e811577bd42",
                "timestamp": 1661955123341
            }
        }


    Response

    .. code-block:: json

        {
            "id": "734235c2-13d2-4574-be68-723e818c08f3",
            "status": 200,
            "result": [{
                "symbol": "BTCUSDT",
                "orderId": 12569099453,
                "orderListId": -1,
                "clientOrderId": "4d96324ff9d44481926157",
                "price": "23416.10000000",
                "origQty": "0.00847000",
                "executedQty": "0.00847000",
                "cummulativeQuoteQty": "198.33521500",
                "status": "FILLED",
                "timeInForce": "GTC",
                "type": "LIMIT",
                "side": "SELL",
                "stopPrice": "0.00000000",
                "icebergQty": "0.00000000",
                "time": 1660801715639,
                "updateTime": 1660801717945,
                "isWorking": true,
                "workingTime": 1660801715639,
                "origQuoteOrderQty": "0.00000000",
                "selfTradePreventionMode": "NONE",
                "preventedMatchId": 0,
                "preventedQuantity": "1.200000"
                }],
            "rateLimits": [{
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 10
            }]
        }

    """

    parameters = purge_map(kwargs)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "allOrders",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)


def oco_history(self, **kwargs):
    """Account OCO history (USER_DATA)

    Keyword Args:
        id (str, optional): Client generated ID.
        fromId (int, optional): OCO ID
        startTime (int, optional): Timestamp in ms
        endTime (int, optional): Timestamp in ms
        limit (int, optional): Default 500; max 1000.
        recvWindow (int, optional): The value cannot be greater than 60000

    Message sent:

    .. code-block:: json

        {
            "id": "8617b7b3-1b3d-4dec-94cd-eefd929b8ceb",
            "method": "allOrderLists",
            "params": {
                "startTime": 1660780800000,
                "endTime": 1660867200000,
                "limit": 5,
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "c8e1484db4a4a02d0e84dfa627eb9b8298f07ebf12fcc4eaf86e4a565b2712c2",
                "timestamp": 1661955123341
            }
        }


    Response

    .. code-block:: json

        {
            "id": "8617b7b3-1b3d-4dec-94cd-eefd929b8ceb",
            "status": 200,
            "result": [{
                "orderListId": 1274512,
                "contingencyType": "OCO",
                "listStatusType": "EXEC_STARTED",
                "listOrderStatus": "EXECUTING",
                "listClientOrderId": "08985fedd9ea2cf6b28996",
                "transactionTime": 1660801713793,
                "symbol": "BTCUSDT",
                "orders": [{
                    "symbol": "BTCUSDT",
                    "orderId": 12569138901,
                    "clientOrderId": "BqtFCj5odMoWtSqGk2X9tU"
                },
                {
                    "symbol": "BTCUSDT",
                    "orderId": 12569138902,
                    "clientOrderId": "jLnZpj5enfMXTuhKB1d0us"
                }]
            }],
            "rateLimits": [{
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 10
            }]
        }

    """

    parameters = purge_map(kwargs)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "allOrderLists",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)


def my_trades(self, **kwargs):
    """Account trade history (USER_DATA)

    Keyword Args:
        id (str, optional): Client generated ID.
        symbol (str, optional): Symbol
        orderId (int, optional): order id
        startTime (int, optional): Timestamp in ms
        endTime (int, optional): Timestamp in ms
        fromId (int, optional): Trade id to fetch from. Default gets most recent trades.
        limit (int, optional): Default 500; max 1000.
        recvWindow (int, optional): The value cannot be greater than 60000

    Message sent:

    .. code-block:: json

        {
            "id": "f4ce6a53-a29d-4f70-823b-4ab59391d6e8",
            "method": "myTrades",
            "params": {
                "symbol": "BTCUSDT",
                "startTime": 1660780800000,
                "endTime": 1660867200000,
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "c5a5ffb79fd4f2e10a92f895d488943a57954edf5933bde3338dfb6ea6d6eefc",
                "timestamp": 1661955125250
            }
        }


    Response:

    .. code-block:: json

        {
            "id": "f4ce6a53-a29d-4f70-823b-4ab59391d6e8",
            "status": 200,
            "result": [
                {
                    "symbol": "BTCUSDT",
                    "id": 12569138901,
                    "orderId": 12569138901,
                    "price": "0.00000000",
                    "qty": "0.00000000",
                    "commission": "0.00000000",
                    "commissionAsset": "BTC",
                    "time": 1660801715639,
                    "isBuyer": true,
                    "isMaker": false,
                    "isBestMatch": true
                }
            ],
            "rateLimits": [{
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 10
            }]
        }

    """

    parameters = purge_map(kwargs)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "myTrades",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)


def prevented_matches(self, symbol: str, **kwargs):
    """Account prevented matches (USER_DATA)

    Keyword Args:
        id (str, optional): Client generated ID.
        symbol (str, optional): Symbol
        preventedMatchId (int, optional): Prevented match id
        orderId (int, optional): order id
        fromPrevMatchId (int, optional): Prevented match id to fetch from.
        limit (int, optional): Default 500; max 1000.
        recvWindow (int, optional): The value cannot be greater than 60000


    Combinations supported::
      - symbol + preventedMatchId
      - symbol + orderId
      - symbol + orderId + fromPreventedMatchId (limit will default to 500)
      - symbol + orderId + fromPreventedMatchId + limit


    Message sent:

    .. code-block:: json

        {
            "id": "g4ce6a53-a39d-4f71-823b-4ab5r391d6y8",
            "method": "myPreventedMatches",
            "params": {
                "symbol": "BTCUSDT",
                "orderId": 35,
                "apiKey": "vmPUZE6mv9SD5VNHk4HlWFsOr6aKE2zvsw0MuIgwCIPy6utIco14y7Ju91duEh8A",
                "signature": "c5a5ffb79fd4f2e10a92f895d488943a57954edf5933bde3338dfb6ea6d6eefc",
                "timestamp": 1673923281052
            }
        }


    Response

    .. code-block:: json

        {
            "id": "g4ce6a53-a39d-4f71-823b-4ab5r391d6y8",
            "status": 200,
            "result": [{
                "symbol": "BTCUSDT",
                "preventedMatchId": 1,
                "takerOrderId": 5,
                "makerOrderId": 3,
                "tradeGroupId": 1,
                "selfTradePreventionMode": "EXPIRE_MAKER",
                "price": "1.100000",
                "makerPreventedQuantity": "1.300000",
                "transactTime": 1669101687094
            }],
            "rateLimits": [{
                "rateLimitType": "REQUEST_WEIGHT",
                "interval": "MINUTE",
                "intervalNum": 1,
                "limit": 1200,
                "count": 10
            }]
        }

    """
    parameters = {"symbol": symbol, **kwargs}
    parameters = purge_map(parameters)

    payload = {
        "id": parameters.pop("id", get_uuid()),
        "method": "myPreventedMatches",
        "params": websocket_api_signature(self.api_key, self.api_secret, parameters),
    }

    self.send(payload)



from binance.lib.utils import check_required_parameter
from binance.lib.utils import check_required_parameters


def new_order_test(self, symbol: str, side: str, type: str, **kwargs):
    """Test New Order (TRADE)

    Test new order creation and signature/recvWindow. Creates and validates a new order but does not send it into
    the matching engine.

    POST /api/v3/order/test

    https://binance-docs.github.io/apidocs/spot/en/#test-new-order-trade

    Args:
        symbol (str)
        side (str)
        type (str)
    Keyword Args:
        timeInForce (str, optional)
        quantity (float, optional)
        quoteOrderQty (float, optional)
        price (float, optional)
        newClientOrderId (str, optional): A unique id among open orders. Automatically generated if not sent.
        stopPrice (float, optional): Used with STOP_LOSS, STOP_LOSS_LIMIT, TAKE_PROFIT, and TAKE_PROFIT_LIMIT orders.
        icebergQty (float, optional): Used with LIMIT, STOP_LOSS_LIMIT, and TAKE_PROFIT_LIMIT to create an iceberg order.
        newOrderRespType (str, optional): Set the response JSON. ACK, RESULT, or FULL;
                MARKET and LIMIT order types default to FULL, all other orders default to ACK.
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameters([[symbol, "symbol"], [side, "side"], [type, "type"]])
    params = {"symbol": symbol, "side": side, "type": type, **kwargs}
    url_path = "/api/v3/order/test"
    return self.sign_request("POST", url_path, params)


def new_order(self, symbol: str, side: str, type: str, **kwargs):
    """New Order (TRADE)

    Post a new order

    POST /api/v3/order

    https://binance-docs.github.io/apidocs/spot/en/#new-order-trade

    Args:
        symbol (str)
        side (str)
        type (str)
    Keyword Args:
        timeInForce (str, optional)
        quantity (float, optional)
        quoteOrderQty (float, optional)
        price (float, optional)
        newClientOrderId (str, optional): A unique id among open orders. Automatically generated if not sent.
        strategyId (int, optional)
        strategyType (int, optional): The value cannot be less than 1000000.
        stopPrice (float, optional): Used with STOP_LOSS, STOP_LOSS_LIMIT, TAKE_PROFIT, and TAKE_PROFIT_LIMIT orders.
        icebergQty (float, optional): Used with LIMIT, STOP_LOSS_LIMIT, and TAKE_PROFIT_LIMIT to create an iceberg order.
        newOrderRespType (str, optional): Set the response JSON. ACK, RESULT, or FULL;
                MARKET and LIMIT order types default to FULL, all other orders default to ACK.
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameters([[symbol, "symbol"], [side, "side"], [type, "type"]])
    params = {"symbol": symbol, "side": side, "type": type, **kwargs}
    url_path = "/api/v3/order"
    return self.sign_request("POST", url_path, params)


def cancel_order(self, symbol: str, **kwargs):
    """Cancel Order (TRADE)

    Cancel an active order.

    DELETE /api/v3/order

    https://binance-docs.github.io/apidocs/spot/en/#cancel-order-trade

    Args:
        symbol (str)
    Keyword Args:
        orderId (int, optional)
        origClientOrderId (str, optional)
        newClientOrderId (str, optional)
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameter(symbol, "symbol")

    url_path = "/api/v3/order"
    payload = {"symbol": symbol, **kwargs}
    return self.sign_request("DELETE", url_path, payload)


def cancel_open_orders(self, symbol: str, **kwargs):
    """Cancel all Open Orders on a Symbol (TRADE)

    Cancels all active orders on a symbol.
    This includes OCO orders.

    DELETE api/v3/openOrders

    https://binance-docs.github.io/apidocs/spot/en/#cancel-all-open-orders-on-a-symbol-trade

    Args:
        symbol (str)
    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameter(symbol, "symbol")

    url_path = "/api/v3/openOrders"
    payload = {"symbol": symbol, **kwargs}
    return self.sign_request("DELETE", url_path, payload)


def get_order(self, symbol, **kwargs):
    """Query Order (USER_DATA)

    Check an order's status.

    GET /api/v3/order

    https://binance-docs.github.io/apidocs/spot/en/#query-order-user_data

    Args:
        symbol (str)
    Keyword Args:
        orderId (int, optional)
        origClientOrderId (str, optional)
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameter(symbol, "symbol")

    url_path = "/api/v3/order"
    payload = {"symbol": symbol, **kwargs}
    return self.sign_request("GET", url_path, payload)


def cancel_and_replace(
    self, symbol: str, side: str, type: str, cancelReplaceMode: str, **kwargs
):
    """Cancel an Existing Order and Send a New Order (USER_DATA)

    Cancels an existing order and places a new order on the same symbol.

    Filters are evaluated before the cancel order is placed.

    If the new order placement is successfully sent to the engine, the order count will increase by 1.

    Weight(IP): 1

    POST /api/v3/order/cancelReplace

    https://binance-docs.github.io/apidocs/spot/en/#cancel-an-existing-order-and-send-a-new-order-user_data

    Args:
        symbol (str)
        side (str)
        type (str)
        cancelReplaceMode (str)
    Keyword Args:
        timeInForce (str, optional): Order time in force
        quantity (float, optional): Order quantity
        quoteOrderQty (float, optional): Quote quantity
        price (float, optional): Order price
        cancelNewClientOrderId (str, optional): Used to uniquely identify this cancel. Automatically generated by default
        cancelOrigClientOrderId (str, optional): Either the cancelOrigClientOrderId or cancelOrderId must be provided. If both are provided, cancelOrderId takes precedence.
        cancelOrderId (int, optional): Either the cancelOrigClientOrderId or cancelOrderId must be provided. If both are provided, cancelOrderId takes precedence.
        newClientOrderId (str, optional): Used to identify the new order. Automatically generated by default
        strategyId (int, optional)
        strategyType (int, optional): The value cannot be less than 1000000.
        stopPrice (float, optional): Used with STOP_LOSS, STOP_LOSS_LIMIT, TAKE_PROFIT, and TAKE_PROFIT_LIMIT orders.
        trailingDelta (float, optional): Used with STOP_LOSS, STOP_LOSS_LIMIT, TAKE_PROFIT, and TAKE_PROFIT_LIMIT orders.
        icebergQty (float, optional): Used with LIMIT, STOP_LOSS_LIMIT, and TAKE_PROFIT_LIMIT to create an iceberg order.
        newOrderRespType (str, optional): Set the response JSON. MARKET and LIMIT order types default to FULL, all other orders default to ACK.
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameters(
        [
            [symbol, "symbol"],
            [side, "side"],
            [type, "type"],
            [cancelReplaceMode, "cancelReplaceMode"],
        ]
    )

    params = {
        "symbol": symbol,
        "side": side,
        "type": type,
        "cancelReplaceMode": cancelReplaceMode,
        **kwargs,
    }
    url_path = "/api/v3/order/cancelReplace"
    return self.sign_request("POST", url_path, params)


def get_open_orders(self, symbol=None, **kwargs):
    """Current Open Orders (USER_DATA)

    Get all open orders on a symbol.

    GET /api/v3/openOrders

    https://binance-docs.github.io/apidocs/spot/en/#current-open-orders-user_data

    Args:
        symbol (str, optional)
    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    url_path = "/api/v3/openOrders"
    payload = {"symbol": symbol, **kwargs}
    return self.sign_request("GET", url_path, payload)


def get_orders(self, symbol: str, **kwargs):
    """All Orders (USER_DATA)

    Get all account orders; active, canceled, or filled.

    GET /api/v3/allOrders

    https://binance-docs.github.io/apidocs/spot/en/#all-orders-user_data

    Args:
        symbol (str)
    Keyword Args:
        orderId (int, optional)
        startTime (int, optional)
        endTime (int, optional)
        limit (int, optional): Default 500; max 1000.
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameter(symbol, "symbol")

    url_path = "/api/v3/allOrders"
    payload = {"symbol": symbol, **kwargs}
    return self.sign_request("GET", url_path, payload)


def new_oco_order(
    self,
    symbol: str,
    side: str,
    quantity: float,
    price: float,
    stopPrice: float,
    **kwargs
):
    """New OCO (TRADE)

    Post a new oco order

    POST /api/v3/order/oco

    https://binance-docs.github.io/apidocs/spot/en/#new-oco-trade

    Args:
        symbol (str)
        side (str)
        quantity (float)
        price (float)
        stopPrice (float)
    Keyword Args:
        listClientOrderId (str, optional): A unique Id for the entire orderList
        limitClientOrderId (str, optional)
        limitStrategyId (int, optional)
        limitStrategyType (int, optional): The value cannot be less than 1000000.
        limitIcebergQty (float, optional)
        stopClientOrderId (str, optional)
        stopStrategyId (int, optional)
        stopStrategyType (int, optional): The value cannot be less than 1000000.
        stopLimitPrice (float, optional)
        stopIcebergQty (float, optional)
        stopLimitTimeInForce (str, optional)
        newOrderRespType (str, optional): Set the response JSON.
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameters(
        [
            [symbol, "symbol"],
            [side, "side"],
            [quantity, "quantity"],
            [price, "price"],
            [stopPrice, "stopPrice"],
        ]
    )
    params = {
        "symbol": symbol,
        "side": side,
        "quantity": quantity,
        "price": price,
        "stopPrice": stopPrice,
        **kwargs,
    }

    url_path = "/api/v3/order/oco"
    return self.sign_request("POST", url_path, params)


def cancel_oco_order(self, symbol, **kwargs):
    """Cancel OCO (TRADE)

    Cancel an entire Order List

    DELETE /api/v3/orderList

    https://binance-docs.github.io/apidocs/spot/en/#cancel-oco-trade

    Args:
        symbol (str)
    Keyword Args:
        orderListId (int, optional): Either orderListId or listClientOrderId must be provided
        listClientOrderId (str, optional): Either orderListId or listClientOrderId must be provided
        newClientOrderId (str, optional): Used to uniquely identify this cancel. Automatically generated by default.
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameter(symbol, "symbol")

    url_path = "/api/v3/orderList"
    payload = {"symbol": symbol, **kwargs}
    return self.sign_request("DELETE", url_path, payload)


def get_oco_order(self, **kwargs):
    """Query OCO (USER_DATA)

    Retrieves a specific OCO based on provided optional parameters

    GET /api/v3/orderList

    https://binance-docs.github.io/apidocs/spot/en/#query-oco-user_data

    Keyword Args:
        orderListId (int, optional): Either orderListId or listClientOrderId must be provided
        origClientOrderId (str, optional): Either orderListId or listClientOrderId must be provided.
        recvWindow (int, optional): The value cannot be greater than 60000
    """
    url_path = "/api/v3/orderList"
    return self.sign_request("GET", url_path, {**kwargs})


def get_oco_orders(self, **kwargs):
    """Query all OCO (USER_DATA)

    Retrieves all OCO based on provided optional parameters

    GET /api/v3/allOrderList

    https://binance-docs.github.io/apidocs/spot/en/#query-all-oco-user_data

    Keyword Args:
        fromId (int, optional): If supplied, neither startTime or endTime can be provided
        startTime (int, optional)
        endTime (int, optional)
        limit (int, optional): Default Value: 500; Max Value: 1000
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    url_path = "/api/v3/allOrderList"
    return self.sign_request("GET", url_path, {**kwargs})


def get_oco_open_orders(self, **kwargs):
    """Query Open OCO (USER_DATA)

    GET /api/v3/openOrderList

    https://binance-docs.github.io/apidocs/spot/en/#query-open-oco-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    url_path = "/api/v3/openOrderList"
    return self.sign_request("GET", url_path, {**kwargs})


def account(self, **kwargs):
    """Account Information (USER_DATA)

    Get current account information

    GET /api/v3/account

    https://binance-docs.github.io/apidocs/spot/en/#account-information-user_data

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    url_path = "/api/v3/account"
    return self.sign_request("GET", url_path, {**kwargs})


def my_trades(self, symbol: str, **kwargs):
    """Account Trade List (USER_DATA)

    Get trades for a specific account and symbol.

    GET /api/v3/myTrades

    https://binance-docs.github.io/apidocs/spot/en/#account-trade-list-user_data

    Args:
        symbol (str)
    Keyword Args:
        fromId (int, optional): TradeId to fetch from. Default gets most recent trades.
        orderId (int, optional): This can only be used in combination with symbol
        startTime (int, optional)
        endTime (int, optional)
        limit (int, optional): Default Value: 500; Max Value: 1000
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    check_required_parameter(symbol, "symbol")

    url_path = "/api/v3/myTrades"
    payload = {"symbol": symbol, **kwargs}
    return self.sign_request("GET", url_path, payload)


def get_order_rate_limit(self, **kwargs):
    """Query Current Order Count Usage (TRADE)

    Displays the user's current order count usage for all intervals.

    GET /api/v3/rateLimit/order

    https://binance-docs.github.io/apidocs/spot/en/#query-current-order-count-usage-trade

    Keyword Args:
        recvWindow (int, optional): The value cannot be greater than 60000
    """

    url_path = "/api/v3/rateLimit/order"
    return self.sign_request("GET", url_path, {**kwargs})



from binance.api import API


class Spot(API):
    def __init__(self, api_key=None, api_secret=None, **kwargs):
        if "base_url" not in kwargs:
            kwargs["base_url"] = "https://api.binance.com"
        super().__init__(api_key, api_secret, **kwargs)

    # MARKETS
    from binance.spot._market import ping
    from binance.spot._market import time
    from binance.spot._market import exchange_info
    from binance.spot._market import depth
    from binance.spot._market import trades
    from binance.spot._market import historical_trades
    from binance.spot._market import agg_trades
    from binance.spot._market import klines
    from binance.spot._market import ui_klines
    from binance.spot._market import avg_price
    from binance.spot._market import ticker_24hr
    from binance.spot._market import ticker_price
    from binance.spot._market import book_ticker
    from binance.spot._market import rolling_window_ticker

    # ACCOUNT (including orders and trades)
    from binance.spot._trade import new_order_test
    from binance.spot._trade import new_order
    from binance.spot._trade import cancel_order
    from binance.spot._trade import cancel_open_orders
    from binance.spot._trade import get_order
    from binance.spot._trade import cancel_and_replace
    from binance.spot._trade import get_open_orders
    from binance.spot._trade import get_orders
    from binance.spot._trade import new_oco_order
    from binance.spot._trade import cancel_oco_order
    from binance.spot._trade import get_oco_order
    from binance.spot._trade import get_oco_orders
    from binance.spot._trade import get_oco_open_orders
    from binance.spot._trade import account
    from binance.spot._trade import my_trades
    from binance.spot._trade import get_order_rate_limit

    # STREAMS
    from binance.spot._data_stream import new_listen_key
    from binance.spot._data_stream import renew_listen_key
    from binance.spot._data_stream import close_listen_key
    from binance.spot._data_stream import new_margin_listen_key
    from binance.spot._data_stream import renew_margin_listen_key
    from binance.spot._data_stream import close_margin_listen_key
    from binance.spot._data_stream import new_isolated_margin_listen_key
    from binance.spot._data_stream import renew_isolated_margin_listen_key
    from binance.spot._data_stream import close_isolated_margin_listen_key

    # MARGIN
    from binance.spot._margin import margin_transfer
    from binance.spot._margin import margin_borrow
    from binance.spot._margin import margin_repay
    from binance.spot._margin import margin_asset
    from binance.spot._margin import margin_pair
    from binance.spot._margin import margin_all_assets
    from binance.spot._margin import margin_all_pairs
    from binance.spot._margin import margin_pair_index
    from binance.spot._margin import new_margin_order
    from binance.spot._margin import cancel_margin_order
    from binance.spot._margin import margin_transfer_history
    from binance.spot._margin import margin_load_record
    from binance.spot._margin import margin_repay_record
    from binance.spot._margin import margin_interest_history
    from binance.spot._margin import margin_force_liquidation_record
    from binance.spot._margin import margin_account
    from binance.spot._margin import margin_order
    from binance.spot._margin import margin_open_orders
    from binance.spot._margin import margin_open_orders_cancellation
    from binance.spot._margin import margin_all_orders
    from binance.spot._margin import margin_my_trades
    from binance.spot._margin import margin_max_borrowable
    from binance.spot._margin import margin_max_transferable
    from binance.spot._margin import isolated_margin_transfer
    from binance.spot._margin import isolated_margin_transfer_history
    from binance.spot._margin import isolated_margin_account
    from binance.spot._margin import isolated_margin_pair
    from binance.spot._margin import isolated_margin_all_pairs
    from binance.spot._margin import toggle_bnbBurn
    from binance.spot._margin import bnbBurn_status
    from binance.spot._margin import margin_interest_rate_history
    from binance.spot._margin import new_margin_oco_order
    from binance.spot._margin import cancel_margin_oco_order
    from binance.spot._margin import get_margin_oco_order
    from binance.spot._margin import get_margin_oco_orders
    from binance.spot._margin import get_margin_open_oco_orders
    from binance.spot._margin import cancel_isolated_margin_account
    from binance.spot._margin import enable_isolated_margin_account
    from binance.spot._margin import isolated_margin_account_limit
    from binance.spot._margin import margin_fee
    from binance.spot._margin import isolated_margin_fee
    from binance.spot._margin import isolated_margin_tier
    from binance.spot._margin import margin_order_usage
    from binance.spot._margin import margin_dust_log
    from binance.spot._margin import summary_of_margin_account

    # SAVINGS
    from binance.spot._savings import savings_flexible_products
    from binance.spot._savings import savings_flexible_user_left_quota
    from binance.spot._savings import savings_purchase_flexible_product
    from binance.spot._savings import savings_flexible_user_redemption_quota
    from binance.spot._savings import savings_flexible_redeem
    from binance.spot._savings import savings_flexible_product_position
    from binance.spot._savings import savings_project_list
    from binance.spot._savings import savings_purchase_project
    from binance.spot._savings import savings_project_position
    from binance.spot._savings import savings_account
    from binance.spot._savings import savings_purchase_record
    from binance.spot._savings import savings_redemption_record
    from binance.spot._savings import savings_interest_history
    from binance.spot._savings import savings_change_position

    # Staking
    from binance.spot._staking import staking_product_list
    from binance.spot._staking import staking_purchase_product
    from binance.spot._staking import staking_redeem_product
    from binance.spot._staking import staking_product_position
    from binance.spot._staking import staking_history
    from binance.spot._staking import staking_set_auto_staking
    from binance.spot._staking import staking_product_quota

    # WALLET
    from binance.spot._wallet import system_status
    from binance.spot._wallet import coin_info
    from binance.spot._wallet import account_snapshot
    from binance.spot._wallet import disable_fast_withdraw
    from binance.spot._wallet import enable_fast_withdraw
    from binance.spot._wallet import withdraw
    from binance.spot._wallet import deposit_history
    from binance.spot._wallet import withdraw_history
    from binance.spot._wallet import deposit_address
    from binance.spot._wallet import account_status
    from binance.spot._wallet import api_trading_status
    from binance.spot._wallet import dust_log
    from binance.spot._wallet import user_universal_transfer
    from binance.spot._wallet import user_universal_transfer_history
    from binance.spot._wallet import transfer_dust
    from binance.spot._wallet import asset_dividend_record
    from binance.spot._wallet import asset_detail
    from binance.spot._wallet import trade_fee
    from binance.spot._wallet import funding_wallet
    from binance.spot._wallet import user_asset
    from binance.spot._wallet import api_key_permissions
    from binance.spot._wallet import bnb_convertible_assets
    from binance.spot._wallet import convertible_coins
    from binance.spot._wallet import toggle_auto_convertion
    from binance.spot._wallet import cloud_mining_trans_history
    from binance.spot._wallet import convert_transfer
    from binance.spot._wallet import convert_history

    # MINING
    from binance.spot._mining import mining_algo_list
    from binance.spot._mining import mining_coin_list
    from binance.spot._mining import mining_worker
    from binance.spot._mining import mining_worker_list
    from binance.spot._mining import mining_earnings_list
    from binance.spot._mining import mining_bonus_list
    from binance.spot._mining import mining_statistics_list
    from binance.spot._mining import mining_account_list
    from binance.spot._mining import mining_hashrate_resale_request
    from binance.spot._mining import mining_hashrate_resale_cancellation
    from binance.spot._mining import mining_hashrate_resale_list
    from binance.spot._mining import mining_hashrate_resale_details
    from binance.spot._mining import mining_account_earning

    # SUB-ACCOUNT
    from binance.spot._sub_account import sub_account_create
    from binance.spot._sub_account import sub_account_list
    from binance.spot._sub_account import sub_account_assets
    from binance.spot._sub_account import sub_account_deposit_address
    from binance.spot._sub_account import sub_account_deposit_history
    from binance.spot._sub_account import sub_account_status
    from binance.spot._sub_account import sub_account_enable_margin
    from binance.spot._sub_account import sub_account_margin_account
    from binance.spot._sub_account import sub_account_margin_account_summary
    from binance.spot._sub_account import sub_account_enable_futures
    from binance.spot._sub_account import sub_account_futures_transfer
    from binance.spot._sub_account import sub_account_margin_transfer
    from binance.spot._sub_account import sub_account_transfer_to_sub
    from binance.spot._sub_account import sub_account_transfer_to_master
    from binance.spot._sub_account import sub_account_transfer_sub_account_history
    from binance.spot._sub_account import sub_account_futures_asset_transfer_history
    from binance.spot._sub_account import sub_account_futures_asset_transfer
    from binance.spot._sub_account import sub_account_spot_summary
    from binance.spot._sub_account import sub_account_universal_transfer
    from binance.spot._sub_account import sub_account_universal_transfer_history
    from binance.spot._sub_account import sub_account_futures_account
    from binance.spot._sub_account import sub_account_futures_account_summary
    from binance.spot._sub_account import sub_account_futures_position_risk
    from binance.spot._sub_account import sub_account_spot_transfer_history
    from binance.spot._sub_account import sub_account_enable_leverage_token
    from binance.spot._sub_account import managed_sub_account_deposit
    from binance.spot._sub_account import managed_sub_account_assets
    from binance.spot._sub_account import managed_sub_account_withdraw
    from binance.spot._sub_account import sub_account_update_ip_restriction
    from binance.spot._sub_account import sub_account_api_get_ip_restriction
    from binance.spot._sub_account import sub_account_api_delete_ip
    from binance.spot._sub_account import managed_sub_account_get_snapshot
    from binance.spot._sub_account import managed_sub_account_investor_trans_log
    from binance.spot._sub_account import managed_sub_account_trading_trans_log
    from binance.spot._sub_account import managed_sub_account_deposit_address

    # FUTURES
    from binance.spot._futures import futures_transfer
    from binance.spot._futures import futures_transfer_history
    from binance.spot._futures import futures_loan_borrow_history
    from binance.spot._futures import futures_loan_repay_history
    from binance.spot._futures import futures_loan_wallet
    from binance.spot._futures import futures_loan_adjust_collateral_history
    from binance.spot._futures import futures_loan_liquidation_history
    from binance.spot._futures import futures_loan_interest_history

    # BLVTs
    from binance.spot._blvt import blvt_info
    from binance.spot._blvt import subscribe_blvt
    from binance.spot._blvt import subscription_record
    from binance.spot._blvt import redeem_blvt
    from binance.spot._blvt import redemption_record
    from binance.spot._blvt import user_limit_info

    # BSwap
    from binance.spot._bswap import bswap_pools
    from binance.spot._bswap import bswap_liquidity
    from binance.spot._bswap import bswap_liquidity_add
    from binance.spot._bswap import bswap_liquidity_remove
    from binance.spot._bswap import bswap_liquidity_operation_record
    from binance.spot._bswap import bswap_request_quote
    from binance.spot._bswap import bswap_swap
    from binance.spot._bswap import bswap_swap_history
    from binance.spot._bswap import bswap_pool_configure
    from binance.spot._bswap import bswap_add_liquidity_preview
    from binance.spot._bswap import bswap_remove_liquidity_preview
    from binance.spot._bswap import bswap_unclaimed_rewards
    from binance.spot._bswap import bswap_claim_rewards
    from binance.spot._bswap import bswap_claimed_rewards

    # FIAT
    from binance.spot._fiat import fiat_order_history
    from binance.spot._fiat import fiat_payment_history

    # C2C
    from binance.spot._c2c import c2c_trade_history

    # LOANS
    from binance.spot._loan import loan_history
    from binance.spot._loan import loan_borrow
    from binance.spot._loan import loan_borrow_history
    from binance.spot._loan import loan_ongoing_orders
    from binance.spot._loan import loan_repay
    from binance.spot._loan import loan_repay_history
    from binance.spot._loan import loan_adjust_ltv
    from binance.spot._loan import loan_adjust_ltv_history
    from binance.spot._loan import loan_vip_ongoing_orders
    from binance.spot._loan import loan_vip_repay
    from binance.spot._loan import loan_vip_repay_history
    from binance.spot._loan import loan_vip_collateral_account
    from binance.spot._loan import loan_loanable_data
    from binance.spot._loan import loan_collateral_data
    from binance.spot._loan import loan_collateral_rate
    from binance.spot._loan import loan_customize_margin_call

    # PAY
    from binance.spot._pay import pay_history

    # CONVERT
    from binance.spot._convert import convert_trade_history

    # REBATE
    from binance.spot._rebate import rebate_spot_history

    # NFT
    from binance.spot._nft import nft_transaction_history
    from binance.spot._nft import nft_deposit_history
    from binance.spot._nft import nft_withdraw_history
    from binance.spot._nft import nft_asset

    # Gift Card (Binance Code in the API documentation)
    from binance.spot._gift_card import gift_card_create_code
    from binance.spot._gift_card import gift_card_redeem_code
    from binance.spot._gift_card import gift_card_verify_code
    from binance.spot._gift_card import gift_card_rsa_public_key
    from binance.spot._gift_card import gift_card_buy_code
    from binance.spot._gift_card import gift_card_token_limit

    # Portfolio Margin
    from binance.spot._portfolio_margin import portfolio_margin_account
    from binance.spot._portfolio_margin import portfolio_margin_collateral_rate
    from binance.spot._portfolio_margin import portfolio_margin_bankruptcy_loan_amount
    from binance.spot._portfolio_margin import portfolio_margin_bankruptcy_loan_repay



from enum import Enum, auto


class AutoName(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name


class TransferType(AutoName):
    MAIN_UMFUTURE = auto()
    MAIN_CMFUTURE = auto()
    MAIN_MARGIN = auto()
    UMFUTURE_MAIN = auto()
    UMFUTURE_MARGIN = auto()
    CMFUTURE_MAIN = auto()
    CMFUTURE_MARGIN = auto()
    MARGIN_MAIN = auto()
    MARGIN_UMFUTURE = auto()
    MARGIN_CMFUTURE = auto()
    ISOLATEDMARGIN_MARGIN = auto()
    MARGIN_ISOLATEDMARGIN = auto()
    ISOLATEDMARGIN_ISOLATEDMARGIN = auto()
    MAIN_FUNDING = auto()
    FUNDING_MAIN = auto()
    FUNDING_UMFUTURE = auto()
    UMFUTURE_FUNDING = auto()
    MARGIN_FUNDING = auto()
    FUNDING_MARGIN = auto()
    FUNDING_CMFUTURE = auto()
    CMFUTURE_FUNDING = auto()



import json
import logging
from binance.lib.utils import get_timestamp
from binance.websocket.binance_socket_manager import BinanceSocketManager


class BinanceWebsocketClient:
    ACTION_SUBSCRIBE = "SUBSCRIBE"
    ACTION_UNSUBSCRIBE = "UNSUBSCRIBE"

    def __init__(
        self,
        stream_url,
        on_message=None,
        on_open=None,
        on_close=None,
        on_error=None,
        on_ping=None,
        on_pong=None,
        logger=None,
    ):
        if not logger:
            logger = logging.getLogger(__name__)
        self.logger = logger
        self.socket_manager = self._initialize_socket(
            stream_url,
            on_message,
            on_open,
            on_close,
            on_error,
            on_ping,
            on_pong,
            logger,
        )

        # start the thread
        self.socket_manager.start()
        self.logger.debug("Binance WebSocket Client started.")

    def _initialize_socket(
        self,
        stream_url,
        on_message,
        on_open,
        on_close,
        on_error,
        on_ping,
        on_pong,
        logger,
    ):
        return BinanceSocketManager(
            stream_url,
            on_message=on_message,
            on_open=on_open,
            on_close=on_close,
            on_error=on_error,
            on_ping=on_ping,
            on_pong=on_pong,
            logger=logger,
        )

    def _single_stream(self, stream):
        if isinstance(stream, str):
            return True
        elif isinstance(stream, list):
            return False
        else:
            raise ValueError("Invalid stream name, expect string or array")

    def send(self, message: dict):
        self.socket_manager.send_message(json.dumps(message))

    def send_message_to_server(self, message, action=None, id=None):
        if not id:
            id = get_timestamp()

        if action != self.ACTION_UNSUBSCRIBE:
            return self.subscribe(message, id=id)
        return self.unsubscribe(message, id=id)

    def subscribe(self, stream, id=None):
        if not id:
            id = get_timestamp()
        if self._single_stream(stream):
            stream = [stream]
        json_msg = json.dumps({"method": "SUBSCRIBE", "params": stream, "id": id})
        self.socket_manager.send_message(json_msg)

    def unsubscribe(self, stream, id=None):
        if not id:
            id = get_timestamp()
        if self._single_stream(stream):
            stream = [stream]
        json_msg = json.dumps({"method": "UNSUBSCRIBE", "params": stream, "id": id})
        self.socket_manager.send_message(json_msg)

    def ping(self):
        self.logger.debug("Sending ping to Binance WebSocket Server")
        self.socket_manager.ping()

    def stop(self, id=None):
        self.socket_manager.close()
        self.socket_manager.join()

    def list_subscribe(self, id=None):
        """sending the list subscription message, e.g.

        {"method": "LIST_SUBSCRIPTIONS","id": 100}

        """

        if not id:
            id = get_timestamp()
        self.socket_manager.send_message(
            json.dumps({"method": "LIST_SUBSCRIPTIONS", "id": id})
        )



from binance.lib.utils import check_required_parameter


def nft_transaction_history(self, orderType: int, **kwargs):
    """Get NFT Transaction History (USER_DATA)

    GET /sapi/v1/nft/history/transactions

    https://binance-docs.github.io/apidocs/spot/en/#get-nft-transaction-history-user_data

    Args:
      orderType (int): 0: purchase order, 1: sell order, 2: royalty income, 3: primary market order, 4: mint fee
    Keyword Args:
      startTime (int, optional)
      endTime (int, optional)
      limit (int, optional): Default 50, Max 50
      page (int, optional): Default 1
      recvWindow (int, optional): The value cannot be greater than 60000
    """
    check_required_parameter(orderType, "orderType")
    params = {"orderType": orderType, **kwargs}
    return self.sign_request("GET", "/sapi/v1/nft/history/transactions", params)


def nft_deposit_history(self, **kwargs):
    """Get NFT Deposit History(USER_DATA)

    GET /sapi/v1/nft/history/deposit

    https://binance-docs.github.io/apidocs/spot/en/#get-nft-deposit-history-user_data

    Keyword Args:
      startTime (int, optional)
      endTime (int, optional)
      limit (int, optional): Default 50, Max 50
      page (int, optional): Default 1
      recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/nft/history/deposit", kwargs)


def nft_withdraw_history(self, **kwargs):
    """Get NFT Withdraw History (USER_DATA)

    GET /sapi/v1/nft/history/withdraw

    https://binance-docs.github.io/apidocs/spot/en/#get-nft-withdraw-history-user_data

    Keyword Args:
      startTime (int, optional)
      endTime (int, optional)
      limit (int, optional): Default 50, Max 50
      page (int, optional): Default 1
      recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/nft/history/withdraw", kwargs)


def nft_asset(self, **kwargs):
    """Get NFT Asset (USER_DATA)

    GET /sapi/v1/nft/user/getAsset

    https://binance-docs.github.io/apidocs/spot/en/#get-nft-asset-user_data

    Keyword Args:
      limit (int, optional): Default 50, Max 50
      page (int, optional): Default 1
      recvWindow (int, optional): The value cannot be greater than 60000
    """

    return self.sign_request("GET", "/sapi/v1/nft/user/getAsset", kwargs)
